from .boston_housing import BostonHousingData
from .concrete import ConcreteData
from .energy_efficiency import EnergyEfficiencyData
from .kin8nm import Kin8nmData
from .naval_propulsion import NavalPropulsionData
from .ccpp import CCPPData
from .protein_structure import ProteinStructureData
from .red_wine import RedWineData
from .yacht_hydrodynamics import YachtHydrodynamicsData
from .year_prediction_msd import YearPredictionMSDData
from .openml import MnistData, FashionMnistData, Cifar10, SVHN


DATASETS = {
    'boston_housing': BostonHousingData,
    'concrete': ConcreteData,
    'energy_efficiency': EnergyEfficiencyData,
    'kin8nm': Kin8nmData,
    'naval_propulsion': NavalPropulsionData,
    'ccpp': CCPPData,
    'protein_structure': ProteinStructureData,
    'red_wine': RedWineData,
    'yacht_hydrodynamics': YachtHydrodynamicsData,
    'year_prediction_msd': YearPredictionMSDData,
    'mnist': MnistData,
    'fashion_mnist': FashionMnistData,
    'cifar_10': Cifar10,
    'svhn': SVHN
}


def build_dataset(name, **kwargs):
    return DATASETS[name](**kwargs)



import numpy as np


class EagerSampleSelector:
    """
    Move samples from pool dataset to train dataset
    Takes samples with highest uncertainty first
    """
    def update_sets(self, X_train, y_train, X_pool, ue_values, sample_size, oracle):
        """
        Update X_train and X_pool by choosing points with highest uncertainty estimations.
        :param X_train, y_train: the datasets that should be extended
        :param ue_values: uncertainty estimation for points in x_pool
        :param sample_size: how much points to add
        :param oracle: the class that should generate y_values for new samples
        :return:
        """
        # obtain new values from pool
        indices = self.sample(ue_values, sample_size)
        X_selected = X_pool[indices]
        Y_selected = oracle.evaluate(X_pool, indices)

        # change sets
        X_train = np.concatenate([X_train, X_selected])
        y_train = np.concatenate([y_train, Y_selected])
        X_pool = np.delete(X_pool, indices, axis=0)

        return X_train, y_train, X_pool

    @staticmethod
    def sample(uncertainty, sample_size):
        return np.argsort(uncertainty)[::-1][:sample_size]


class StochasticSampleSelector(EagerSampleSelector):
        """
        Move samples from pool dataset to train dataset
        Takes samples with probability propotional to uncertainty
        """
        @staticmethod
        def sample(uncertainty, sample_size):
            indexes = np.arange(len(uncertainty))
            uncertainty = uncertainty.astype('double')
            uncertainty[uncertainty<0] = 0
            probabilities = uncertainty / np.sum(uncertainty)
            return np.random.choice(indexes, sample_size, replace=False, p=probabilities)



import numpy as np
import torch
import torch.nn.functional as F
from dppy.finite_dpps import FiniteDPP


from alpaca.uncertainty_estimator import build_estimator
from alpaca.uncertainty_estimator.masks import build_mask


def update_set(x_pool, x_train, y_pool, y_train, step, method='mc_dropout', model=None, nn_runs=100, task='classification'):
    if task == 'classification':
        samples = torch.FloatTensor(x_pool).cuda()
    else:
        samples = torch.DoubleTensor(x_pool).cuda()

    if method == 'random':
        idxs = range(step)
    elif method == 'AL_dpp':
        mask = build_mask('mc_dropout')
        estimator = build_estimator(
            'bald_masked', model, dropout_mask=mask, num_classes=10,
            keep_runs=True, nn_runs=nn_runs)
        estimator.estimate(samples)  # to generate mcd
        mcd = estimator.last_mcd_runs().reshape(-1, nn_runs * 10)
        dpp = FiniteDPP('likelihood', **{'L': np.corrcoef(mcd)})
        idxs = set()
        while len(idxs) < step:
            dpp.sample_exact()
            idxs.update(dpp.list_of_samples[-1])
        idxs = list(idxs)[:step]
    elif method == 'error_oracle':
        predictions = F.softmax(model(samples), dim=1).detach().cpu().numpy()
        errors = -np.log(predictions[np.arange(len(predictions)),  y_pool])
        idxs = np.argsort(errors)[::-1][:step]
    elif method == 'stoch_oracle':
        predictions = F.softmax(model(samples), dim=1).detach().cpu().numpy()
        errors = -np.log(predictions[np.arange(len(predictions)), y_pool])
        idxs = np.random.choice(len(predictions), step, replace=False, p=errors/sum(errors))
    elif method == 'max_entropy':
        predictions = F.softmax(model(samples), dim=1).detach().cpu().numpy()
        entropies = entropy(predictions)
        idxs = np.argsort(entropies)[::-1][:step]
    else:
        if task == 'classification':
            estimator = build_estimator('bald_masked', model, dropout_mask=method, num_classes=10, nn_runs=nn_runs)
        else:
            estimator = build_estimator('mcdue_masked', model, dropout_mask=method, nn_runs=nn_runs)
        estimations = estimator.estimate(samples)
        idxs = np.argsort(estimations)[::-1][:step]
        estimator.reset()

    x_add, y_add = np.copy(x_pool[idxs]), np.copy(y_pool[idxs])
    x_train = np.concatenate((x_train, x_add))
    y_train = np.concatenate((y_train, y_add))
    x_pool = np.delete(x_pool, idxs, axis=0)
    y_pool = np.delete(y_pool, idxs, axis=0)

    return x_pool, x_train, y_pool, y_train


def entropy(x):
    return -np.sum(x * np.log(np.clip(x, 1e-8, 1)), axis=-1)



import os
import os.path as path

from scipy.optimize import rosen
import numpy as np


class RosenData:
    """
    Generate points for n_dim Rosenbrock function
    Or loads the points from cached version
    """
    def __init__(
            self, n_dim, data_size, data_split=None,
            use_cache=True, cache_dir='dataloader/data/rosen', rewrite=False):

        self.n_dim = n_dim
        self.data_size = data_size
        self.data_split = data_split
        self.cache_dir = cache_dir
        self.splits = self._parse_split(data_split)

        self.x_set, self.y_set = self._build_set(use_cache, rewrite)

    def dataset(self, label):
        assert label in ['train', 'val', 'test', 'pool']
        index = self.splits[label]
        return self.x_set[index[0]:index[1]], self.y_set[index[0]:index[1]]

    def _build_set(self, use_cache, rewrite):
        if use_cache:
            try:
                return self._load()
            except AssertionError:
                print("Can't use cache, generating new dataset")

        x = 2*np.random.random((self.data_size, self.n_dim)) - 1
        y = rosen(x.T)[:, None]
        if rewrite:
            self._save('x', x)
            self._save('y', y)
        return x, y

    def _save(self, name, dataset):
        if not(path.exists(self.cache_dir)):
            os.makedirs(self.cache_dir)
        np.save(path.join(self.cache_dir, name + '.npy'), dataset)

    def _load(self):
        x_path = path.join(self.cache_dir, 'x.npy')
        y_path = path.join(self.cache_dir, 'y.npy')
        assert path.exists(x_path) and path.exists(y_path)
        x_set = np.load(x_path)
        y_set = np.load(y_path)
        assert x_set.shape == (self.data_size, self.n_dim)

        return x_set, y_set

    def _parse_split(self, splits):
        assert abs(sum(splits) - 1) < 1e-6  # split should sum to 1
        indices = [round(self.data_size*sum(splits[:i])) for i in range(4)]

        return {
            'train': (indices[0], indices[1]),
            'val': (indices[1], indices[2]),
            'test': (indices[2], indices[3]),
            'pool': (indices[3], self.data_size),
        }



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from alpaca.dataloader.custom_dataset import loader


class BaseMLP(nn.Module):
    def __init__(self, layer_sizes, activation, postprocessing=lambda x: x, device=None):
        super(BaseMLP, self).__init__()

        if device is None:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            self.device = device

        self.layer_sizes = layer_sizes
        self.fcs = []
        for i, layer in enumerate(layer_sizes[:-1]):
            fc = nn.Linear(layer, layer_sizes[i+1])
            setattr(self, 'fc'+str(i), fc)  # to register params
            self.fcs.append(fc)
        self.postprocessing = postprocessing
        self.activation = activation
        self.double()
        self.to(self.device)

    def forward(self, x, dropout_rate=0, dropout_mask=None):
        x = self.activation(self.fcs[0](x))

        for layer_num, fc in enumerate(self.fcs[1:-1]):
            x = self.activation(fc(x))
            if dropout_mask is None:
                x = nn.Dropout(dropout_rate)(x)
            else:
                x = x*dropout_mask(x, dropout_rate, layer_num)
        x = self.fcs[-1](x)
        x = self.postprocessing(x)
        return x

    def fit(
            self, train_set, val_set, epochs=10000, verbose=True,
            validation_step=100, patience=5, batch_size=500, dropout_rate=0):
        train_loader = loader(*train_set, batch_size=batch_size, shuffle=True)

        best_val_loss = float('inf')
        current_patience = patience

        # Train the model
        for epoch in range(epochs):
            for points, labels in train_loader:
                # Move tensors to the configured device
                points = points.reshape(-1, self.layer_sizes[0]).to(self.device)
                labels = labels.to(self.device)

                # Forward pass
                outputs = self(points, dropout_rate=dropout_rate)
                loss = self.criterion(outputs, labels)

                # Backward and optimize
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
            # Print intermediate results and check patience
            if (epoch + 1) % validation_step == 0:
                val_loss = self.evaluate(val_set)
                if verbose:
                    self._print_status(epoch, epochs, loss.item(), val_loss)
                if val_loss < best_val_loss:
                    best_val_loss = val_loss
                    current_patience = patience
                else:
                    current_patience -= 1
                    if current_patience <= 0:
                        break
        self.val_loss = val_loss

    def evaluate(self, dataset, y_scaler=None):
        """ Return model losses for provided data loader """
        data_loader = loader(*dataset)
        with torch.no_grad():
            losses = []
            for points, labels in data_loader:
                points = points.reshape(-1, self.layer_sizes[0]).to(self.device)
                labels = labels.to(self.device)
                outputs = self(points)
                if y_scaler is not None:
                    outputs = torch.Tensor(y_scaler.inverse_transform(outputs.cpu()))
                    labels = torch.Tensor(y_scaler.inverse_transform(labels.cpu()))
                losses.append(self.criterion(outputs, labels).item())

        return sum(losses)/len(losses)

    def _print_status(self, epoch, epochs, loss, val_loss):
        print('Epoch [{}/{}], Loss: {:.4f}, Validation loss: {:.4f}'
              .format(epoch + 1, epochs, loss, val_loss))


class MLP(BaseMLP):
    def __init__(self, layer_sizes, postprocessing=None, loss=nn.MSELoss,
                 optimizer=None, activation=None, **kwargs):
        if postprocessing is None:
            postprocessing = lambda x: x

        if activation is None:
            activation = F.celu

        super(MLP, self).__init__(layer_sizes, activation=activation, postprocessing=postprocessing, **kwargs)

        self.criterion = loss()

        if optimizer is None:
            optimizer = {'type': 'Adadelta'}
        self.optimizer = self.init_optimizer(optimizer)

    def init_optimizer(self, optimizer):
        if isinstance(optimizer, dict):
            kwargs = optimizer.copy()
            opt_type = getattr(torch.optim, kwargs.pop('type'))
            kwargs['params'] = self.parameters()
            optimizer = opt_type(**kwargs)
        elif not isinstance(optimizer, torch.optim.Optimizer):
            raise TypeError(
                'optimizer must be either an Optimizer object or a dict, '
                'but got {}'.format(type(optimizer)))
        return optimizer




from os import path

import pandas as pd
import numpy as np

from .config import DATA_DIR
from .saver import DataSaver
from .downloader import download
from sklearn.model_selection import train_test_split


URL = 'https://archive.ics.uci.edu/ml/machine-learning-databases/00242/ENB2012_data.xlsx'


# TODO: use both output features
class EnergyEfficiencyData:
    """Load/provides boston housing dataset"""
    def __init__(self, use_cache=False, val_split=0.2):
        self.use_cache = use_cache
        cache_dir = path.join(DATA_DIR, 'dataloader/data/energy_efficiency')
        self.saver = DataSaver(cache_dir)
        self.val_split = val_split
        self._build_dataset(cache_dir)

    def dataset(self, label):
        if self.use_cache:
            return self.saver.load(label)

        data = self.data[label]

        x, y = data[:, :-2], data[:, -1:]
        self.saver.save(x, y, label)
        return x, y

    def _build_dataset(self, cache_dir):
        data_path = download(cache_dir, 'energy_efficiencty.xlsx', URL)
        self.df = pd.read_excel(data_path)
        table = self.df.to_numpy()
        if self.val_split != 0:
            train, val = train_test_split(table, test_size=self.val_split, shuffle=True)
        else:
            train, val = table, []
        self.data = {'train': train, 'val': val}




from .nngp import NNGPRegression
from .mcdue import MCDUE, MCDUEMasked
from .eue import EnsembleUE, EnsembleMCDUE, EnsembleNLLUE, EnsembleMCDUEMasked
from .random_estimator import RandomEstimator
from .bald import Bald, BaldMasked, BaldEnsemble


def build_estimator(name, model, **kwargs):
    if name == 'nngp':
        estimator = NNGPRegression(model, **kwargs)
    elif name == 'random':
        estimator = RandomEstimator()
    elif name == 'mcdue':
        estimator = MCDUE(model, **kwargs)
    elif name == 'mcdue_masked':
        estimator = MCDUEMasked(model, **kwargs)
    elif name == 'bald':
        estimator = Bald(model, **kwargs)
    elif name == 'bald_masked':
        estimator = BaldMasked(model, **kwargs)
    elif name == 'bald_ensemble':
        estimator = BaldEnsemble(model, **kwargs)
    elif name == 'eue_nll':
        estimator = EnsembleNLLUE(model)
    elif name == 'eue':
        estimator = EnsembleUE(model)
    elif name == 'emcdue':
        estimator = EnsembleMCDUE(model, **kwargs)
    elif name == 'emcdue_masked':
        estimator = EnsembleMCDUEMasked(model, **kwargs)
    else:
        raise ValueError("Wrong estimator name")
    return estimator



import torch
import torch.nn as nn
import torch.nn.functional as F


class SimpleConv(nn.Module):
    def __init__(self, num_classes=10, activation=None):
        if activation is None:
            self.activation = F.leaky_relu
        else:
            self.activation = activation
        super().__init__()
        self.conv1 = nn.Conv2d(1, 16, 3)
        self.conv2 = nn.Conv2d(16, 32, 3)
        self.linear_size = 12*12*32
        self.fc1 = nn.Linear(self.linear_size, 256)
        self.dropout = nn.Dropout(0.5)
        self.fc2 = nn.Linear(256, 10)

    def forward(self, x, dropout_rate=0., dropout_mask=None):
        x = self.activation(self.conv1(x))
        x = self.activation(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, self.linear_size)
        x = self.activation(self.fc1(x))
        x = self._dropout(x, dropout_mask, dropout_rate, 0)
        x = self.fc2(x)
        return x

    def _dropout(self, x, dropout_mask, dropout_rate, layer_num):
        if dropout_mask is None:
            x = self.dropout(x)
        else:
            x = x * dropout_mask(x, dropout_rate, layer_num)
        return x


class StrongConv(nn.Module):
    def __init__(self, num_classes=10, activation=None):
        super().__init__()
        if activation is None:
            self.activation = F.leaky_relu
        else:
            self.activation = activation
        self.conv11 = nn.Conv2d(3, 16, 3, padding=1)
        self.conv12 = nn.Conv2d(16, 16, 3, padding=1)
        self.dropout = nn.Dropout(0.25)

        self.conv21 = nn.Conv2d(16, 32, 3, padding=1)
        self.conv22 = nn.Conv2d(32, 32, 3, padding=1)

        self.linear_size = 8*8*32
        self.fc1 = nn.Linear(self.linear_size, 256)
        self.fc2 = nn.Linear(256, num_classes)

    def forward(self, x, dropout_rate=0., dropout_mask=None):
        x = F.relu(self.conv11(x))
        x = F.relu(self.conv12(x))
        x = F.max_pool2d(x, 2, 2)
        if dropout_mask is None:
            x = self.dropout(x)

        x = F.relu(self.conv21(x))
        x = F.relu(self.conv22(x))
        x = F.max_pool2d(x, 2, 2)
        if dropout_mask is None:
            x = self.dropout(x)
        x = x.view(-1, self.linear_size)
        x = F.relu(self.fc1(x))
        x = self._dropout(x, dropout_mask, dropout_rate, 1)
        x = self.fc2(x)
        return x

    def _dropout(self, x, dropout_mask, dropout_rate, layer_num):
        if dropout_mask is None:
            x = self.dropout(x)
        else:
            x = x * dropout_mask(x, dropout_rate, layer_num)
        return x


class MediumConv(nn.Module):
    def __init__(self, num_classes=10):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 16, 3)
        self.conv2 = nn.Conv2d(16, 32, 3)
        self.linear_size = 14*14*32
        self.fc1 = nn.Linear(self.linear_size, 256)
        self.dropout = nn.Dropout(0.5)
        self.fc2 = nn.Linear(256, num_classes)

    def forward(self, x, dropout_rate=0., dropout_mask=None):
        x = F.elu(self.conv1(x))
        x = F.elu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, self.linear_size)
        # x = self._dropout(x, dropout_mask, dropout_rate, 0)
        x = F.elu(self.fc1(x))
        x = self._dropout(x, dropout_mask, dropout_rate, 1)
        x = self.fc2(x)
        return x

    def _dropout(self, x, dropout_mask, dropout_rate, layer_num):
        if dropout_mask is None:
            x = self.dropout(x)
        else:
            x = x * dropout_mask(x, dropout_rate, layer_num)
        return x


class AnotherConv(nn.Module):
    def __init__(self, num_classes=10, activation=None):
        if activation is None:
            self.activation = F.celu
        else:
            self.activation = activation
        super().__init__()
        self.conv1 = nn.Conv2d(3, 16, 3)
        self.conv2 = nn.Conv2d(16, 32, 3)
        self.linear_size = 14*14*32
        self.fc1 = nn.Linear(self.linear_size, 256)
        self.dropout = nn.Dropout(0.5)
        self.fc2 = nn.Linear(256, 10)

    def forward(self, x, dropout_rate=0.5, dropout_mask=None):
        x = self.activation(self.conv1(x))
        x = self.activation(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, self.linear_size)
        x = self.activation(self.fc1(x))
        x = self._dropout(x, dropout_mask, dropout_rate, 1)
        x = self.fc2(x)
        return x

    def _dropout(self, x, dropout_mask, dropout_rate, layer_num):
        if dropout_mask is None:
            x = self.dropout(x)
        else:
            x = x * dropout_mask(x, dropout_rate, layer_num)
        return x



import os
from os import path

import requests


def download(cache_dir, name, url):
    if not path.exists(cache_dir):
        os.makedirs(cache_dir)
    data_path = path.join(cache_dir, name)
    if not path.exists(data_path):
        print("Loading dataset...")
        response = requests.get(url)
        with open(data_path, 'wb') as f:
            f.write(response.content)

    return data_path


import numpy as np


class IdentityOracle:
    """
    Basic oracle

    Getting points for y suppose to be expensive operation
    But for this oracle it's not and it's mostly placeholder
    for more advanced oracle implementations
    """
    def __init__(self, y_set):
        self.y_set = y_set

    def evaluate(self, x_set, indices):
        """x_set and y_set are expected to be numpy ndarrays"""
        y_selected = self.y_set[indices]
        self.y_set = np.delete(self.y_set, indices, axis=0)
        return y_selected


import numpy as np
from numpy.random import uniform, normal
from .saver import DataSaver


class ToyQubicData:
    """Generate noisy points of x*3 on [-4, 4]"""
    def __init__(self, noise=9, use_cache=False):
        self.noise = noise
        self.use_cache = use_cache
        self.saver = DataSaver('dataloader/data/toy_qubic')

    def dataset(self, label):
        points = 20
        x0, x1 = -4, 4
        k = 1.5

        if self.use_cache:
            return self.saver.load(label)
        if label in ['train', 'val']:
            x = uniform(x0, x1, points)
            y = np.power(x, 3) + normal(0, self.noise, points)
        elif label == 'ground_truth':
            x = np.arange(k*x0, k*x1, 0.2)
            y = np.power(x, 3)

        x = np.expand_dims(x, axis=1)
        y = np.expand_dims(y, axis=1)
        self.saver.save(x, y, label)
        return x, y


class ToySinData:
    def __init__(self, noise=0.02, use_cache=False):
        self.noise = noise
        self.use_cache = use_cache
        self.saver = DataSaver('dataloader/data/toy_sin')

    def dataset(self, label):
        points = 30

        if self.use_cache:
            return self.saver.load(label)
        if label in ['train', 'val']:
            x = np.concatenate((uniform(-1, -0.3, points), uniform(0.5, 1.2, points)))
            # x = np.concatenate((uniform(-4, -3.3, points), uniform(3.2, 4, points)))
            y = self._function(x) + normal(0, self.noise, 2*points)
        elif label == 'ground_truth':
            x = np.arange(-1.5, 1.5, 0.04)
            # x = np.arange(-5, 5, 0.04)
            y = self._function(x)

        x = np.expand_dims(x, axis=1)
        y = np.expand_dims(y, axis=1)
        self.saver.save(x, y, label)
        return x, y

    def _function(self, x):
        return 0.3*x + 0.3*np.sin(2*np.pi*x) + 0.3*np.sin(4*np.pi*x)






import numpy as np
import torch
import torch.nn.functional as F
from torch.utils.data import TensorDataset, DataLoader


class Trainer:
    def __init__(
            self, model, batch_size=128, lr=1e-3, dropout_train=0.5, weight_decay=1e-4,
            loss=None, regression=False):
        self.model = model
        self.device = 'cuda'
        self.model.to(self.device)

        # self.optimizer = torch.optim.Adam(model.parameters(), lr=lr)
        self.optimizer = torch.optim.Adadelta(model.parameters(), weight_decay=weight_decay)
        self.dropout_train = dropout_train
        self.batch_size = batch_size

        self.loss = loss or F.cross_entropy
        self.regression = regression

        self.val_loss_history = []
        self.train_loss_history = []

    def fit(self, train_set, val_set, epochs=10, log_interval=1000, verbose=False, patience=5, dropout_rate=None):
        if dropout_rate is None:
            dropout_rate = self.dropout_train

        self.model.train()
        loader = self._to_loader(*train_set)
        val_loader = self._to_loader(*val_set)

        self._set_patience(patience)

        for epoch in range(epochs):
            for batch_idx, (data, target) in enumerate(loader):
                data, target = data.to(self.device), target.to(self.device)
                self.optimizer.zero_grad()
                output = self.model(data, dropout_rate=dropout_rate)
                loss = self.loss(output, target)
                loss.backward()
                self.optimizer.step()

                if verbose:
                    self._report(epoch, loss, data, batch_idx, log_interval, loader, val_loader)

            if not self._check_patience(val_loader):
                break

    def _set_patience(self, patience):
        self.start_patience = patience
        self.current_patience = self.start_patience
        self.best_loss = float('inf')

    def _check_patience(self, val_loader):
        loss = self.evaluate(val_loader)
        if loss < self.best_loss:
            self.best_loss = loss
            self.current_patience = self.start_patience
        else:
            self.current_patience -= 1

        return self.current_patience > 0

    def _report(self, epoch, loss, data, batch_idx, log_interval, loader, val_loader):
        if (batch_idx + 1) % log_interval == 0 or batch_idx == len(loader) - 1:
            val_loss = self.evaluate(val_loader)
            percent = batch_idx / len(loader)
            print("Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tVal Loss: {:.6f}".format(
                epoch, int(percent*len(loader.dataset)), len(loader.dataset), 100 * percent,
                loss.item(), val_loss))

            self.val_loss_history.append(val_loss)
            self.train_loss_history.append(loss)

    def evaluate(self, val_loader):
        if not isinstance(val_loader, DataLoader):
            if len(val_loader) == 1:
                val_loader = self._to_loader(val_loader)
            elif len(val_loader) == 2:
                val_loader = self._to_loader(val_loader[0], val_loader[1])

        losses = []
        with torch.no_grad():
            for data, target in val_loader:
                data, target = data.to(self.device), target.to(self.device)
                output = self.model(data, dropout_rate=0)
                losses.append(self.loss(output, target).item())
        return np.mean(losses)

    def predict(self, x, logits=False):
        self.model.eval()
        predictions = []
        with torch.no_grad():
            loader = self._to_loader(x, shuffle=False)
            for batch in loader:
                batch = batch[0].to(self.device)
                prediction = self.model(batch)
                if not logits and not self.regression:
                    prediction = prediction.argmax(dim=1, keepdim=True)
                predictions.append(prediction.cpu())
            predictions = torch.cat(predictions).numpy()
        return predictions

    def _to_loader(self, x, y=None, shuffle=True):
        if y is None:
            ds = TensorDataset(torch.FloatTensor(x))
        else:
            if self.regression:
                ds = TensorDataset(torch.FloatTensor(x), torch.FloatTensor(y))
            else:
                ds = TensorDataset(torch.FloatTensor(x), torch.LongTensor(y.reshape(-1)))
        loader = DataLoader(ds, batch_size=self.batch_size, shuffle=shuffle)
        return loader

    def __call__(self, x, dropout_rate=0., dropout_mask=None):
        x = torch.FloatTensor(x).to(self.device)
        return self.model(x, dropout_rate=dropout_rate, dropout_mask=dropout_mask)

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()


class EnsembleTrainer:
    def __init__(
            self, model_class, model_kwargs, n_models, batch_size=128,
            dropout_train=0.5, reduction='mean'):
        self.models = [model_class(**model_kwargs) for _ in range(n_models)]
        self.trainers = [
            Trainer(model, batch_size=batch_size, dropout_train=dropout_train)
            for model in self.models]
        self.reduction = reduction
        self.device = 'cuda'

    def fit(self, train_set, val_set, **kwargs):
        for trainer in self.trainers:
            trainer.fit(train_set, val_set, **kwargs)

    def predict(self, x):
        self.eval()
        with torch.no_grad():
            # predictions = self(x).detach().argmax(dim=1, keepdim=True)
            predictions = self(x).argmax(axis=1)[..., np.newaxis]
        return predictions.cpu().numpy()

    def evaluate(self, val_loader):
        return np.mean([trainer.evaluate(val_loader) for trainer in self.trainers])

    def __call__(self, x, reduction='default', **kwargs):
        res = torch.stack([torch.Tensor(trainer.predict(x, logits=True)) for trainer in self.trainers])

        if reduction == 'default':
            reduction = self.reduction

        if reduction is None:
            res = res
        elif reduction == 'mean':
            res = res.mean(dim=0)

        return res

    def train(self):
        [trainer.train() for trainer in self.trainers]

    def eval(self):
        [trainer.eval() for trainer in self.trainers]



import torch
from scipy.special import softmax
import numpy as np
from .masks import build_mask


class Bald:
    """
    Estimate uncertainty in classification tasks for samples with MCDUE approach
    """
    def __init__(self, net, nn_runs=25, dropout_rate=.5, num_classes=1):
        self.net = net
        self.nn_runs = nn_runs
        self.dropout_rate = dropout_rate
        self.num_classes = num_classes

    def estimate(self, X_pool, *args):
        mcd_runs = np.zeros((X_pool.shape[0], self.nn_runs, self.num_classes))

        with torch.no_grad():
            for nn_run in range(self.nn_runs):
                prediction = self.net(X_pool, dropout_rate=self.dropout_rate)
                mcd_runs[:, nn_run] = prediction.to('cpu')

        return _bald(mcd_runs)


class BaldMasked:
    """
    Estimate uncertainty for samples with MCDUE approach
    """
    # TODO: different acquisition to separate classes, it's not BALD
    def __init__(
            self, net, nn_runs=100, dropout_mask=None, dropout_rate=.5,
            num_classes=2, keep_runs=False, acquisition='var_ratio'):
        self.net = net
        self.nn_runs = nn_runs
        self.num_classes = num_classes
        self.dropout_rate = dropout_rate
        if isinstance(dropout_mask, str):
            dropout_mask = build_mask(dropout_mask)
        self.dropout_mask = dropout_mask
        self.keep_runs = keep_runs
        self._mcd_runs = np.array([])
        self.acquisition = acquisition

    def estimate(self, X_pool, *args):
        mcd_runs = np.zeros((X_pool.shape[0], self.nn_runs, self.num_classes))

        with torch.no_grad():
            self.net.eval()
            # Some mask needs first run without dropout, i.e. decorrelation mask
            if hasattr(self.dropout_mask, 'dry_run') and self.dropout_mask.dry_run:
                self.net(X_pool, dropout_rate=self.dropout_rate, dropout_mask=self.dropout_mask)

            # Get mcdue estimation
            for nn_run in range(self.nn_runs):
                prediction = self.net(
                    X_pool, dropout_rate=self.dropout_rate, dropout_mask=self.dropout_mask
                ).to('cpu')
                mcd_runs[:, nn_run] = prediction

            if self.keep_runs:
                self._mcd_runs = mcd_runs

        return self._aquisition(mcd_runs)

    def _aquisition(self, mcd_runs):
        if self.acquisition == 'var_ratio':
            predictions = np.argmax(mcd_runs, axis=-1)
            # count how many time repeats the strongest class
            mode_count = lambda preds : np.max(np.bincount(preds))
            modes = [mode_count(point) for point in predictions]
            ue = 1 - np.array(modes) / self.nn_runs
            return ue
        elif self.acquisition == 'var_soft':
            probabilities = softmax(mcd_runs, axis=-1)
            ue = np.mean(np.std(probabilities, axis=-2), axis=-1)
            return ue
        elif self.acquisition == 'bald':
            return _bald(mcd_runs)
        else:
            raise ValueError

    def reset(self):
        if hasattr(self.dropout_mask, 'reset'):
            self.dropout_mask.reset()

    def last_mcd_runs(self):
        """Return model prediction for last uncertainty estimation"""
        if not self.keep_runs:
            print("mcd_runs: You should set `keep_runs=True` to properly use this method")
        return self._mcd_runs


class BaldEnsemble:
    """
    Estimate uncertainty in classification tasks for samples with Ensemble approach
    """
    def __init__(self, ensemble, num_classes=1):
        self.ensemble = ensemble
        self.num_classes = num_classes

    def estimate(self, x_pool, *args):
        with torch.no_grad():
            logits = np.array(self.ensemble(x_pool, reduction=None).cpu())

        return _bald(np.swapaxes(logits, 0, 1))


def _entropy(x):
    return np.sum(-x*np.log(np.clip(x, 1e-8, 1)), axis=-1)


def _bald(logits):
    predictions = softmax(logits, axis=-1)

    predictive_entropy = _entropy(np.mean(predictions, axis=1))
    expected_entropy = np.mean(_entropy(predictions), axis=1)

    return predictive_entropy - expected_entropy



from collections import defaultdict

import torch
import numpy as np
import numpy.linalg as la
from scipy.special import softmax
from dppy.finite_dpps import FiniteDPP


DEFAULT_MASKS = ['mc_dropout', 'decorrelating_sc', 'dpp', 'k_dpp']


# It's better to use this function to get the mask then call them directly
def build_masks(names=None, **kwargs):
    masks = {
        'basic_bern': BasicBernoulliMask(),
        'mc_dropout': BasicBernoulliMask(),
        'decorrelating': DecorrelationMask(),
        'decorrelating_sc': DecorrelationMask(scaling=True, dry_run=False),
        'dpp': DPPMask(),
        'k_dpp': KDPPMask(),
        'k_dpp_noisereg': KDPPMask(noise_level=kwargs.get('noise_level', 1e-2))
    }
    if names is None:
        return masks
    return {name: masks[name] for name in names}


# Build single mask
def build_mask(name, **kwargs):
    return build_masks([name], **kwargs)[name]


class BasicBernoulliMask:
    def __call__(self, x, dropout_rate=0.5, layer_num=0):
        p = 1 - dropout_rate

        if p < 0 or p > 1:
            raise ValueError("dropout probability has to be between 0 and 1, "
                             "but got {}".format(p))

        noise = self._make_noise(x)
        if p == 0:
            noise.fill_(0)
        else:
            noise.bernoulli_(p).div_(p)

        noise = noise.expand_as(x)
        return noise

    @staticmethod
    def _make_noise(input):
        return input.new().resize_as_(input)


class DecorrelationMask:
    def __init__(self, scaling=False, dry_run=True):
        self.layer_correlations = {}
        self.scaling = scaling  # use adaptive scaling before softmax
        self.dry_run = dry_run

    def __call__(self, x, dropout_rate=0.5, layer_num=0):
        if layer_num not in self.layer_correlations:
            x_matrix = x.cpu().numpy()
            self.x_matrix = x_matrix

            noise = 1e-8 * np.random.rand(*x_matrix.shape)  # to prevent degeneration
            corrs = np.sum(np.abs(np.corrcoef((x_matrix+noise).T)), axis=1)
            scores = np.reciprocal(corrs)
            if self.scaling:
                scores = 4 * scores / max(scores)
            self.layer_correlations[layer_num] = softmax(scores)
            # Initially we should pass identity mask,
            # otherwise we won't get right correlations for all layers
            return x.data.new(x.data.size()[-1]).fill_(1)
        mask = x.data.new(x.data.size()[-1]).fill_(0)
        k = int(len(mask)*(1-dropout_rate))
        ids = np.random.choice(len(mask), k, p=self.layer_correlations[layer_num])
        mask[ids] = 1 / (1 - dropout_rate + 1e-10)

        return x.data.new(mask)

    def reset(self):
        self.layer_correlations = {}


ATTEMPTS = 30


class DPPMask:
    def __init__(self):
        self.dpps = {}
        self.layer_correlations = {}  # keep for debug purposes
        # Flag for uncertainty estimator to make first run without taking the result
        self.dry_run = True

    def __call__(self, x, dropout_rate=0.5, layer_num=0):
        if layer_num not in self.layer_correlations:
            # warm-up, generatign correlations masks
            x_matrix = x.cpu().numpy()

            self.x_matrix = x_matrix
            micro = 1e-12
            x_matrix += np.random.random(x_matrix.shape) * micro  # for computational stability
            correlations = np.corrcoef(x_matrix.T)
            self.dpps[layer_num] = FiniteDPP('likelihood', **{'L': correlations})
            self.layer_correlations[layer_num] = correlations

            return x.data.new(x.data.size()[-1]).fill_(1)

        # sampling nodes ids
        dpp = self.dpps[layer_num]

        for _ in range(ATTEMPTS):
            dpp.sample_exact()
            ids = dpp.list_of_samples[-1]
            if len(ids):  # We should retry if mask is zero-length
                break

        mask_len = x.data.size()[-1]
        mask = x.data.new(mask_len).fill_(0)
        mask[ids] = mask_len/len(ids)

        return x.data.new(mask)

    def reset(self):
        self.layer_correlations = {}


class KDPPMask:
    def __init__(self, noise_level=None, tol_level=1e-3):
        self.layer_correlations = {}
        self.dry_run = True
        self.dpps = {}
        self.ranks = {}
        self.ranks_history = defaultdict(list)
        self.noise_level = noise_level
        self.tol_level = tol_level

    def _rank(self, dpp):
        N = dpp.eig_vecs.shape[0]
        tol = max(np.max(dpp.L_eig_vals) * N * np.finfo(np.float).eps, self.tol_level)
        rank = np.count_nonzero(dpp.L_eig_vals > tol)
        return rank

    def __call__(self, x, dropout_rate=0.5, layer_num=0):
        if layer_num not in self.layer_correlations:
            x_matrix = x.cpu().numpy()

            correlations = np.corrcoef(x_matrix.T)
            if self.noise_level is not None:
                correlations += self.noise_level * np.eye(len(correlations))
            self.dpps[layer_num] = FiniteDPP('likelihood', **{'L': correlations})
            self.dpps[layer_num].sample_exact()  # to trigger eig values generation
            self.ranks[layer_num] = self._rank(self.dpps[layer_num])

            # Keep data for debugging
            self.ranks_history[layer_num].append(self.ranks[layer_num])
            self.layer_correlations[layer_num] = correlations

            return x.data.new(x.data.size()[-1]).fill_(1)

        mask = x.data.new(x.data.size()[-1]).fill_(0)
        k = int(self.ranks[layer_num] * (1 - dropout_rate))
        self.dpps[layer_num].sample_exact_k_dpp(k)

        ids = self.dpps[layer_num].list_of_samples[-1]

        mask[ids] = 1 / (1 - dropout_rate + 1e-10)

        return x.data.new(mask)

    def reset(self):
        self.layer_correlations = {}



from os import path

import pandas as pd
import numpy as np

from .config import DATA_DIR
from .saver import DataSaver
from .downloader import download
from sklearn.model_selection import train_test_split


URL = 'https://archive.ics.uci.edu/ml/machine-learning-databases/housing/housing.data'


class BostonHousingData:
    """Load/provides boston housing dataset"""
    def __init__(self, use_cache=False, val_split=0.2):
        self.use_cache = use_cache
        cache_dir = path.join(DATA_DIR, 'dataloader/data/boston_housing')
        self.saver = DataSaver(cache_dir)
        self.column_names = [
            'CRIM', 'ZN', 'INDUS', 'CHAS', 'NOX', 'RM', 'AGE',
            'DIS', 'RAD', 'TAX', 'PTRATIO', 'B', 'LSTAT', 'MEDV']
        self.val_split = val_split
        self._build_dataset(cache_dir)

    def dataset(self, label):
        if self.use_cache:
            return self.saver.load(label)
        data = self.data[label]
        x, y = data[:, :-1], data[:, -1:]
        self.saver.save(x, y, label)
        return x, y

    def _build_dataset(self, cache_dir):
        data_path = download(cache_dir, 'housing.data', URL)
        self.df = pd.read_table(data_path, names=self.column_names, header=None, delim_whitespace=True)
        table = self.df.to_numpy()

        if self.val_split != 0:
            train, val = train_test_split(table, test_size=self.val_split, shuffle=True)
        else:
            train, val = table, []

        self.data = {'train': train, 'val': val}




import os.path as path

import pandas as pd
from sklearn.model_selection import train_test_split
import numpy as np

from .config import DATA_DIR
from .downloader import download

URL = 'http://scrippsco2.ucsd.edu/assets/data/atmospheric/stations/in_situ_co2/monthly/monthly_in_situ_co2_mlo.csv'


class CO2Data:
    def __init__(self, val_split=0.2):
        self.cache_dir = path.join(DATA_DIR, 'dataloader/data/co_2', )
        self.val_split = val_split
        self._build_dataset(self.cache_dir)

    def dataset(self, label='train'):
        data = self.data[label]
        x, y = data[:, :1], data[:, 1:]
        return x, y

    def _build_dataset(self, cache_dir):
        data_path = download(cache_dir, 'co2_monthly.csv', URL)
        df = pd.read_csv(data_path, comment='"').drop([0, 1]).iloc[:, [3, 4]]
        for column in df:
            df[column] = df[column].astype(np.float)
        df = df[df.iloc[:, 1] != -99.99]

        table = df.to_numpy()
        if self.val_split == 0:
            train = table
            val = None
        else:
            train, val = train_test_split(table, test_size=self.val_split)

        self.data = {'train': train, 'val': val}


if __name__ == '__main__':
    CO2Data()



import os
from os.path import expanduser
from pathlib import Path


home = expanduser("~")
DATA_DIR = Path(home)/'.local'/'share'/'alpaca'


if not os.path.exists(DATA_DIR):
    os.makedirs(DATA_DIR, exist_ok=True)



from sklearn.preprocessing import StandardScaler
import torch
from torch.utils.data import TensorDataset, DataLoader


def scale(train, val):
    scaler = StandardScaler()
    scaler.fit(train)
    train = scaler.transform(train)
    val = scaler.transform(val)
    return train, val, scaler


# Set initial datas
def loader(x, y, batch_size=128, task='regression'):
    if task == 'regression':
        ds = TensorDataset(torch.DoubleTensor(x), torch.DoubleTensor(y))
    elif task == 'classification':
        ds = TensorDataset(torch.DoubleTensor(x), torch.LongTensor(y))
    _loader = DataLoader(ds, batch_size=batch_size)
    return _loader





import numpy as np
import torch
from .masks import build_mask


class MCDUE:
    """
    Estimate uncertainty for samples with MCDUE approach
    """
    def __init__(self, net, nn_runs=25, dropout_rate=.5):
        self.net = net
        self.nn_runs = nn_runs
        self.dropout_rate = dropout_rate

    def estimate(self, X_pool, *args):
        mcd_realizations = np.zeros((X_pool.shape[0], self.nn_runs))

        with torch.no_grad():
            for nn_run in range(self.nn_runs):
                prediction = self.net(X_pool, dropout_rate=self.dropout_rate)
                mcd_realizations[:, nn_run] = np.ravel(prediction.to('cpu'))

        return np.ravel(np.std(mcd_realizations, axis=1))


class MCDUEMasked:
    """
    Estimate uncertainty for samples with MCDUE approach
    """
    def __init__(self, net, nn_runs=25, dropout_rate=.5, dropout_mask=None, keep_runs=False):
        self.net = net
        self.nn_runs = nn_runs
        self.dropout_rate = dropout_rate
        if isinstance(dropout_mask, str):
            dropout_mask = build_mask(dropout_mask)
        self.dropout_mask = dropout_mask
        self.keep_runs = keep_runs
        self._mcd_runs = np.array([])

    def estimate(self, X_pool, *args):
        mcd_runs = np.zeros((X_pool.shape[0], self.nn_runs))

        with torch.no_grad():
            # Some mask needs first run without dropout, i.e. decorrelation mask
            if hasattr(self.dropout_mask, 'dry_run') and self.dropout_mask.dry_run:
                self.net(X_pool, dropout_rate=self.dropout_rate, dropout_mask=self.dropout_mask)

            # Get mcdue estimation
            for nn_run in range(self.nn_runs):
                prediction = self.net(
                    X_pool, dropout_rate=self.dropout_rate, dropout_mask=self.dropout_mask
                ).to('cpu')
                mcd_runs[:, nn_run] = np.ravel(prediction)

            if self.keep_runs:
                self._mcd_runs = mcd_runs

        return np.ravel(np.std(mcd_runs, axis=1))

    def reset(self):
        if hasattr(self.dropout_mask, 'reset'):
            self.dropout_mask.reset()

    def last_mcd_runs(self):
        """Return model prediction for last uncertainty estimation"""
        if not self.keep_runs:
            print("mcd_runs: You should set `keep_runs=True` to properly use this method")
        return self._mcd_runs








from os import path

import pandas as pd
from sklearn.model_selection import train_test_split

from .config import DATA_DIR
from .saver import DataSaver
from .downloader import download


URL = 'https://archive.ics.uci.edu/ml/machine-learning-databases/00265/CASP.csv'


class ProteinStructureData:
    """Protein structure (CASP) dataset from UCI"""
    def __init__(self, use_cache=False, val_split=0.2):
        self.use_cache = use_cache
        cache_dir = path.join(DATA_DIR, 'dataloader/data/protein_structure')
        self.saver = DataSaver(cache_dir)
        self.val_split = val_split
        self._build_dataset(cache_dir)

    def dataset(self, label):
        if self.use_cache:
            return self.saver.load(label)

        data = self.data[label]
        x, y = data[:, 1:], data[:, :1]
        self.saver.save(x, y, label)
        return x, y

    def _build_dataset(self, cache_dir):
        data_path = download(cache_dir, 'CASP.csv', URL)
        self.df = pd.read_csv(data_path)
        table = self.df.to_numpy()
        if self.val_split != 0:
            train, val = train_test_split(table, test_size=self.val_split, shuffle=True)
        else:
            train, val = table, []
        self.data = {'train': train, 'val': val}


if __name__ == '__main__':
    dataset = ProteinStructureData()
    x_train, y_train = dataset.dataset('train')
    x_val, y_val = dataset.dataset('val')
    print(dataset.df.head())
    print(x_train.shape, y_train.shape, y_val.shape)




from __future__ import print_function

import numpy as np
from matplotlib import pyplot as plt

import matplotlib.colors as colors
import torch
import torch.utils.data
from torch import nn
from torch.autograd import Variable
from torch.nn import functional as F

# from utilities.metric import predictive_entropy
# from utilities.metric import expected_entropy
# from utilities.metric import mutual_information

from tqdm import tqdm


class VAE(nn.Module):
    def __init__(self):
        super(VAE, self).__init__()

        self.encoder1 = nn.Linear(784, 400)
        self.encoder2 = nn.Linear(400, 256)
        self.mu = nn.Linear(256, 2)
        self.logvar = nn.Linear(256, 2)

        self.decoder1 = nn.Linear(2, 256)
        self.decoder2 = nn.Linear(256, 400)
        self.decoder3 = nn.Linear(400, 784)

        self.elu = nn.ELU()
        self.sigmoid = nn.Sigmoid()

    def encode(self, x):
        d1 = self.elu(self.encoder1(x))
        d2 = self.elu(self.encoder2(d1))
        return self.mu(d2), self.logvar(d2)

    def reparameterize(self, mu, logvar):
        if self.training:
            std = logvar.mul(0.5).exp_()
            eps = Variable(std.data.new(std.size()).normal_())
            return eps.mul(std).add_(mu)
        else:
            return mu

    def decode(self, z):
        d1 = self.elu(self.decoder1(z))
        d2 = self.elu(self.decoder2(d1))
        res = self.sigmoid(self.decoder3(d2))
        return res

    def forward(self, x):
        mu, logvar = self.encode(x.view(-1, 784))
        z = self.reparameterize(mu, logvar)
        return self.decode(z), mu, logvar


# Reconstruction + KL divergence losses summed over all elements and batch
def loss_function(recon_x, x, mu, logvar):
    BCE = F.binary_cross_entropy(recon_x, x.view(-1, 784), size_average=False)
    KLD_element = mu.pow(2).add_(logvar.exp()).mul_(-1).add_(1).add_(logvar)
    KLD = torch.sum(KLD_element).mul_(-0.5)

    return BCE + KLD


def train_autoencoder(epoch, train_loader, log_interval, model, optimizer):
    model.train()
    train_loss = 0
    for batch_idx, (data, _) in enumerate(train_loader):
        data = Variable(data)
        data = data.cuda()
        optimizer.zero_grad()
        recon_batch, mu, logvar = model(data)
        loss = loss_function(recon_batch, data.view(-1, 784), mu, logvar)
        loss.backward()
        train_loss += loss.item()
        optimizer.step()
        if batch_idx % log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch,
                batch_idx * len(data),
                len(train_loader.dataset),
                100. * batch_idx / len(train_loader),
                loss.item() / len(data)
            ))

    print('====> Epoch: {} Average loss: {:.4f}'.format(
        epoch, train_loss / len(train_loader.dataset)))


def test_autoencoder(test_loader, model):
    model.eval()
    test_loss = 0
    for i, (data, _) in enumerate(test_loader):
        data = data.cuda()
        data = Variable(data)
        recon_batch, mu, logvar = model(data)
        test_loss += loss_function(
            recon_batch,
            data.view(-1, 1, 28, 28),
            mu,
            logvar
        ).item()

    test_loss /= len(test_loader.dataset)
    print('====> Test set loss: {:.4f}'.format(test_loss))


def show_encoded_decoded_images(images, figsize=(5, 20), cols=5, titles=None):
    assert ((titles is None) or (len(images) == len(titles)))
    n_images = len(images)
    fig = plt.figure(figsize=figsize)

    if titles is None: titles = ['Image (%d)' % i for i in
                                 range(1, n_images + 1)]
    for n, (image, title) in enumerate(zip(images, titles)):

        a = fig.add_subplot(cols, np.ceil(n_images / float(cols)), n + 1)
        if image.ndim == 2:
            plt.gray()
        plt.imshow(image)
        if n == 0:
            plt.title('Original image')
        elif n == 1:
            plt.title('Decoded')
        plt.axis('off')
    plt.show()


def show_decoder_quality(mini_batch, model):
    data = mini_batch.cuda()
    data = Variable(data)
    print(data)
    print(data.shape)
    recon_batch, _, _ = model(data)
    decoded = recon_batch.cpu().data.numpy(
    ).reshape(25, 28, 28)
    to_encode = mini_batch.numpy().reshape(25, 28, 28) #[:, 0, :, :]

    alternated = np.zeros((50, 28, 28))
    alternated[0::2] = to_encode
    alternated[1::2] = decoded

    show_encoded_decoded_images(
        alternated,
        cols=25
    )


def create_latent_space_results_mc(
        mesh_min,
        mesh_max,
        model,
        vae,
        delim=0.1,
        num_inferences=50
):
    latent_X, latent_Y = np.meshgrid(np.arange(mesh_min, mesh_max, delim),
                                     np.arange(mesh_min, mesh_max, delim))
    all_results = []

    row_length = latent_X.shape[0]
    decoded_imgs = []

    for i in tqdm(range(row_length)):
        row_X = latent_X[i]
        row_Y = latent_Y[i]

        t = Variable(torch.Tensor(np.stack([row_X, row_Y]).T).cuda())
        decoded = vae.decode(t).view(-1, 1, 28, 28)
        decoded_imgs.append(decoded.cpu().data.numpy())

        list_ = []

        for j in range(num_inferences):
            list_.append((nn.Softmax(1)(model(decoded))).cpu().data.numpy())

        results = np.stack(list_)

        all_results.append(results)

    p_ents = np.zeros((row_length, row_length))
    e_ents = np.zeros((row_length, row_length))
    m_info = np.zeros((row_length, row_length))

    # for i in range(row_length):
    #     p_ents[i] = predictive_entropy(all_results[i])
    #     e_ents[i] = expected_entropy(all_results[i])
    #     m_info[i] = mutual_information(all_results[i])

    return p_ents, e_ents, m_info, latent_X, latent_Y, decoded_imgs


def create_latent_space_results_is(mesh_min, mesh_max, model, vae, delim=0.1):
    latent_X, latent_Y = np.meshgrid(
        np.arange(mesh_min, mesh_max, delim),
        np.arange(mesh_min, mesh_max, delim)
    )
    all_results = []

    row_length = latent_X.shape[0]
    decoded_imgs = []
    model.eval()

    for i in tqdm(range(row_length)):
        row_X = latent_X[i]
        row_Y = latent_Y[i]

        t = Variable(torch.Tensor(np.stack([row_X, row_Y]).T).cuda())
        decoded = vae.decode(t).view(-1, 1, 28, 28)
        decoded_imgs.append(decoded.cpu().data.numpy())

        all_results.append(
            nn.Softmax(1)(model.forward(decoded)[0])[:, 10].cpu().data.numpy()
        )

    return np.stack(all_results), latent_X, latent_Y, decoded_imgs


def plot_latent_space(
        latent_X,
        latent_Y,
        uncertainty,
        xses,
        yses,
        title,
        alpha_latent=0.8,
        alpha_training=0.05,
        size_latent=80,
        size_training=20,
        norm=True,
        savepath=None
):

    plt.figure(figsize=(15, 15), frameon=False)

    con_yses = yses

    if norm:
        norm_ = colors.LogNorm(vmin=uncertainty.flatten().min(),
                               vmax=uncertainty.flatten().max())
    else:
        norm_ = None

    plt.scatter(
        latent_X.flatten(),
        latent_Y.flatten(),
        s=size_latent,
        alpha=alpha_latent,
        c=uncertainty.flatten(),
        norm=norm_,
        cmap='gray',
        marker='s'

    )

    for i in range(10):
        plt.scatter(
            xses[:, 0][con_yses == i],
            xses[:, 1][con_yses == i],
            s=size_training,
            alpha=alpha_training,
            label=i
        )

    plt.title(title)

    for spine in plt.gca().spines.values():
        spine.set_visible(False)

    if savepath:
        plt.savefig(savepath, bbox_inches = 'tight', pad_inches = 0)


def show_decoded_latent_space(vae, savepath='images/vae_decoded.png'):
    X2, Y2 = np.meshgrid(
        np.arange(-10, 10, 1),
        np.arange(-10, 10, 1),
    )

    input_2 = np.concatenate([X2.reshape(-1, 1), Y2.reshape(-1, 1)], axis=1)
    inf01 = vae.decode(
        Variable(torch.Tensor(input_2)).cuda()).cpu().data.numpy()
    inf_manifold2 = inf01.reshape(20, 20, 28, 28)
    fig = plt.figure(figsize=(15, 15))

    for i in tqdm(range(20)):
        for j in range(20):
            fig.add_subplot(20, 20, i * 20 + j + 1)
            plt.imshow(inf_manifold2[i, j, :, :])
            plt.axis('off')

    plt.savefig(savepath)



from os import path

from sklearn.model_selection import train_test_split
from sklearn.datasets import fetch_openml
import numpy as np

from .config import DATA_DIR
from .saver import DataSaver


class Openml:
    def __init__(self, name, val_size=0.2):
        cache_dir = path.join(DATA_DIR, f'dataloader/data/{name}')
        self.saver = DataSaver(cache_dir)
        self.val_size = val_size
        self.name = name
        self._build_dataset(cache_dir)

    def dataset(self, label):
        return self.data[label]

    def _build_dataset(self, cache_dir):
        if not path.exists(cache_dir):
            x, y = fetch_openml(self.name, return_X_y=True, cache=True)
            y = y.astype(np.int)
            self.saver.save(x, y)
        else:
            x, y = self.saver.load()

        if self.val_size != 0:
            x_train, x_val, y_train, y_val = train_test_split(
                x, y, test_size=self.val_size, shuffle=True, stratify=y)
        else:
            x_train, y_train = x, y
            x_val, y_val = [], []

        self.data = {
            'train': (x_train, y_train),
            'val': (x_val, y_val),
        }


class MnistData(Openml):
    def __init__(self, **kwargs):
        super().__init__('mnist_784', **kwargs)


class FashionMnistData(Openml):
    def __init__(self, **kwargs):
        super().__init__('Fashion-MNIST', **kwargs)


class Cifar10(Openml):
    def __init__(self, **kwargs):
        super().__init__('CIFAR_10', **kwargs)


class SVHN(Openml):
    def __init__(self, **kwargs):
        super().__init__('SVHN', **kwargs)



import os
import os.path as path
import numpy as np


class DataSaver:
    def __init__(self, data_dir):
        self.data_dir = data_dir

    def save(self, x, y, name=''):
        if not(path.exists(self.data_dir)):
            os.makedirs(self.data_dir)
        np.save(path.join(self.data_dir, f"x_{name}.npy"), x)
        np.save(path.join(self.data_dir, f"y_{name}.npy"), y)

    def load(self, name=''):
        x_path = path.join(self.data_dir, f"x_{name}.npy")
        y_path = path.join(self.data_dir, f"y_{name}.npy")
        assert path.exists(x_path) and path.exists(y_path)
        x_set = np.load(x_path, allow_pickle=True)
        y_set = np.load(y_path, allow_pickle=True)
        return x_set, y_set



import torch
import torch.utils.data


def loader(x, y, batch_size=64, shuffle=False):
    custom_dataset = CustomDataset(x, y)

    loader = torch.utils.data.DataLoader(
        dataset=custom_dataset, batch_size=batch_size, shuffle=shuffle)

    return loader


class CustomDataset(torch.utils.data.Dataset):
    def __init__(self, x_set, y_set):
        self.x_set = torch.from_numpy(x_set)
        self.y_set = torch.from_numpy(y_set)

    def __getitem__(self, index):
        return self.x_set[index], self.y_set[index]

    def __len__(self):
        return len(self.x_set)

